/*
   Copyright 2007-2009 Murali Krishnan

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 */

package com.notuvy.singleapp;

import com.notuvy.singleapp.multiuser.GenericEnforcer;
import org.apache.log4j.Logger;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.net.BindException;
import java.net.ServerSocket;
import java.net.Socket;

/**
 * A utility that enforces the invocation of a system-wide singleton application instance.
 *
 * @author  murali
 */
public abstract class SingleUserEnforcer extends GenericEnforcer implements Runnable, Enforcer {

    //------------------------------------------------------------
    //- Class Variables

    protected static final Logger LOG = Logger.getLogger("com.notuvy.singleapp");


    //------------------------------------------------------------
    //- Class Functions


    //------------------------------------------------------------
    //- Instance Variables

    private final int fPort;
    private ServerSocket fServer = null;
    private final Runnable fRunnable;
    private boolean fAlive = true;


    //------------------------------------------------------------
    //- Constructors

    protected SingleUserEnforcer(int pPort, Runnable pRunnable) {
        fPort = pPort;
        fRunnable = pRunnable;
    }


    //------------------------------------------------------------
    //- Accessors

    public int getPort() {
        return fPort;
    }

    private ServerSocket getServer() {
        return fServer;
    }


    //------------------------------------------------------------
    //- Settors

    private void setServer(ServerSocket pServer) {
        fServer = pServer;
    }


    //------------------------------------------------------------
    //- Private/Protected Utility Functions

    protected void takeControl() {
        Thread thread = new Thread(this, "SingleAppEnforcer");
        thread.setDaemon(true);
        thread.start();

        fRunnable.run();
    }

    protected void wrest(Preemptable pPreemptable) {
        try {
            Socket socket = new Socket("localhost", getPort());
            // This should tell the other application to shut down.

            ObjectInputStream ois = new ObjectInputStream(socket.getInputStream());
            boolean hasMessage = ois.readBoolean();
            Object message = hasMessage ? ois.readObject() : null;
            ois.close();
            socket.close();

            pPreemptable.recvFromPreempted(message);

            try {
                Thread.sleep(1000);
                // Wait for port to die.
            } catch (InterruptedException ie) {
                // nothing.
            }

            setServer(new ServerSocket(getPort(), 1));
        } catch (IOException ioe) {
            LOG.error("Problem wresting on [" + getPort() + "].", ioe);
        } catch (ClassNotFoundException cnfe) {
            LOG.error("Undefined class at deserialization.", cnfe);
        }
    }


    //------------------------------------------------------------
    //- Public Interface Functions

    public void attempt() {
        // The first entity to call this will create the socket listener, and thus call "takeControl()".
        // All subsequent calls will encounter a port conflict (until the first application quits), and thus call "handlePortConflict()".
        try {
            setServer(new ServerSocket(getPort(), 1));

            takeControl();
        } catch (BindException be) {
            handlePortConflict();
        } catch (IOException ioe) {
            LOG.error("Could not listen on [" + getPort() + "].", ioe);
        }
    }


    //------------------------------------------------------------
    //- Class Interface Functions

    // Note that this will be executed in its own thread.  See "takeControl()".
    public void run() {
        while (fAlive) {
            try {
                Socket socket = getServer().accept();
                handleConnection(socket);
                socket.close();
            } catch (IOException ioe) {
                // Not an error if not alive.  That means the application was shutdown thus the socket connection was broken.
                if (fAlive) {
                    LOG.error("Problem with a new connection.", ioe);
                }
            }
        }
    }


    protected abstract void handlePortConflict();

    // When the first instance receives a socket connection, handle it.
    protected abstract void handleConnection(Socket pSocket);

    public void shutdownServer() {
        fAlive = false;
        try {
            getServer().close();
        } catch (IOException ioe) {
            LOG.error("Problem shutting down singleapp listener.", ioe);
        }
    }


    //------------------------------------------------------------
    //- Inner Classes

    static class Defer extends SingleUserEnforcer {
        private final Deferable fDeferable;

        protected Defer(int pPort, Deferable pDeferable) {
            super (pPort, pDeferable);
            fDeferable = pDeferable;
        }

        protected void handlePortConflict() {
            deferControl();
        }

        protected void handleConnection(Socket pSocket) {
            LOG.info("Rejected another instance invocation attempt.");

            try {
                ObjectInputStream ois = new ObjectInputStream(pSocket.getInputStream());
                boolean hasMessage = ois.readBoolean();
                Object message = hasMessage ? ois.readObject() : null;
                ois.close();

                fDeferable.recvFromDeferred(message);
            } catch (IOException ioe) {
                LOG.error("Could read forwarded data.", ioe);
            } catch (ClassNotFoundException cnfe) {
                LOG.error("Unexpected class.", cnfe);
            }
        }

        private void deferControl() {
            LOG.error("Not starting; another instance is already running.");
            try {
                Socket socket = new Socket("localhost", getPort());
                Serializable message = fDeferable.sendFromDeferred();
                ObjectOutputStream oos = new ObjectOutputStream(socket.getOutputStream());
                if (message == null) {
                    oos.writeBoolean(false);
                } else {
                    oos.writeBoolean(true);
                    oos.writeObject(message);
                }
                oos.close();
                socket.close();
            } catch (IOException ioe) {
                LOG.error("Problem forwarding on [" + getPort() + "].", ioe);
            }
        }
    }


    static class Preempt extends SingleUserEnforcer {
        private final Preemptable fPreemptable;

        protected Preempt(int pPort, Preemptable pPreemptable, boolean pAllowSystemExit) {
            super (pPort, pPreemptable);
            fPreemptable = pPreemptable;
            fAllowSystemExit = pAllowSystemExit;
        }

        protected void handlePortConflict() {
            wrest(fPreemptable);
            takeControl();
        }

        protected void handleConnection(Socket pSocket) {
            try {
                shutdownServer();  // Stop listening for new connections (there will be none).
//                super.getServer().close();

                Serializable message = fPreemptable.sendToPreemptor();
                ObjectOutputStream oos = new ObjectOutputStream(pSocket.getOutputStream());
                if (message == null) {
                    oos.writeBoolean(false);
                } else {
                    oos.writeBoolean(true);
                    oos.writeObject(message);
                }
                oos.close();
                pSocket.close();
            } catch (IOException ioe) {
                LOG.error("Problem sending to preemptor on [" + getPort() + "].", ioe);
            }

            fPreemptable.gracefulExitForPreempt();
//            System.exit(0);  // Force death if not graceful.
            if (fAllowSystemExit) {
                System.exit(0);
            }
        }
    }



    //------------------------------------------------------------
    //- Main



}
