package com.notuvy.singleapp.multiuser;

import com.notuvy.singleapp.Preemptable;

import java.io.EOFException;
import java.io.IOException;
import java.io.Serializable;
import java.net.Socket;

/**
 * Basic description
 *
 * @author murali
 */
public class MultiUserPreempt extends MultiUserSingleAppEnforcer {

    //------------------------------------------------------------
    //- Class Variables


    //------------------------------------------------------------
    //- Class Functions


    //------------------------------------------------------------
    //- Instance Variables

    private final Preemptable fPreemptable;


    //------------------------------------------------------------
    //- Constructors

    public MultiUserPreempt(int pPort, Preemptable pPreemptable, boolean pAllowSystemExit) {
        super(pPort, pPreemptable);
        fPreemptable = pPreemptable;
        fAllowSystemExit = pAllowSystemExit;
    }


    //------------------------------------------------------------
    //- Gettors


    //------------------------------------------------------------
    //- Settors


    //------------------------------------------------------------
    //- Private/Protected Utility Functions

    private void shutdown() {
        if (getServer() != null) {
            try {
                getServer().close();  // Stop listening for new connections (there will be none).
            } catch (IOException ioe) {
                LOG.error("Problem closing server.", ioe);
            }
        }
        Broker.terminateCurrent();
        fPreemptable.gracefulExitForPreempt();
        if (fAllowSystemExit) {
            System.exit(0);
        }
    }


    //------------------------------------------------------------
    //- Public Interface Functions


    //------------------------------------------------------------
    //- Class Interface Functions

    protected void handlePortConflict() {
        try {
            // This should tell any conflicting instance to shut down.
            BrokerConnection brokerConnection = new BrokerConnection(getUser(), new Socket("localhost", getPort()));

            boolean brokerWasShutdown = false;

            boolean conflict = brokerConnection.readUserConflict(getUser());
            if (conflict) {
                Object message = brokerConnection.readObjectMessage();

                fPreemptable.recvFromPreempted(message);

                brokerWasShutdown = brokerConnection.readOwnedBrokerWasShutdown();
            }

            if (brokerWasShutdown) {
                // If the conflict triggered the broker in another instance to stop, then start a new one here.
                try {
                    Thread.sleep(1000);
                    // Wait to give some time for the port to clear again.
                } catch (InterruptedException ie) {
                    // nothing.
                }

                reattempt();
            } else {
                // Otherwise attach to the broker as a client.
                startBrokerClientListener(brokerConnection, new BrokerClientListener());
            }
            fPreemptable.run();
        } catch (IOException ioe) {
            LOG.error("Problem wresting on [" + getPort() + "].", ioe);
        }
    }

    protected void handleConnection(Socket pSocket) {
        try {
            BrokerConnection incomingConnection = new BrokerConnection(pSocket);

            boolean conflict = alreadyHaveInstanceForUser(incomingConnection);

            incomingConnection.writeUserConflict(conflict);

            if (conflict) {
                boolean shutdown = false;
                Object message;
                BrokerConnection obsoleteConnection = fUserMap.remove(incomingConnection.getUser());
                if (obsoleteConnection == null) {
                    // This is a request to the broker owner.
                    shutdown = true;

                    message = fPreemptable.sendToPreemptor();
                    LOG.warn("Preempted, and shutting down broker (another instance will inherit it).");
                } else {
                    // Forward the request, and get the message to relay.
                    LOG.debug("Forwarding message to preempt another user.");
                    obsoleteConnection.getOutput().writeInt(MSG_TYPE_PREEMPT_MESSAGE);
                    obsoleteConnection.getOutput().flush();
                    message = obsoleteConnection.readObjectMessage();
                    // Now send an ack saying the message was received and the client can continue to shutdown.
                    obsoleteConnection.writeAck();
                    obsoleteConnection.closeByOutput();
                }

                incomingConnection.writeObjectMessage(message);

                incomingConnection.writeOwnedBrokerWasShutdown(shutdown);
                if (shutdown) {
                    shutdown();
                }
            } else {
                LOG.debug("Creating a client connection for another user.");
            }

            // This instance will die (preempted), so replace the client connection.
            fUserMap.put(incomingConnection.getUser(), incomingConnection);
        } catch (IOException ioe) {
            LOG.error("Problem sending to preemptor on [" + getPort() + "].", ioe);
        }
    }

    protected void handleRelayedMessage() throws IOException {
        Serializable message = fPreemptable.sendToPreemptor();
        fClientConnection.writeObjectMessage(message);
    }


    //------------------------------------------------------------
    //- Inner Classes

    class BrokerClientListener implements Runnable {
        public void run() {
            boolean inheritingBroker = false;
            boolean alive = true;
            try {
                LOG.debug("Started listening to broker as client.");
                while (alive) {
                    int msgType = fClientConnection.getInput().readInt();

                    if (MSG_TYPE_PING == msgType) {
                        handlePing();
                    } else if (MSG_TYPE_PREEMPT_MESSAGE == msgType) {
                        handleRelayedMessage();
                        // LOG.error("processed relayed message.");

                        // Need to wait for an ack.  Or else this will continue on and shut down immediatel.
                        // And the message will be lost
                        fClientConnection.readAck();

                        LOG.warn("Preempted by another instance.");
                        alive = false;
                    } else {
                        LOG.error(String.format("Unrecognized message type [%d].", msgType));
                    }
                }
            } catch (EOFException eofe) {
                inheritingBroker = true;
                LOG.debug("Connection to broker lost, this should become broker");
            } catch (IOException ioe) {
                LOG.error("Could not processes forwarded message.", ioe);
            }

            if (inheritingBroker) {
                reattempt();
            } else {
                // LOG.info("Ending the broker client.");
                fPreemptable.gracefulExitForPreempt();
                if (fAllowSystemExit) {
                    System.exit(0);
                }
            }
        }
    }


    //------------------------------------------------------------
    //- Main


}
